from frads import room


def test_make_room():
    width = 3
    depth = 5
    floor_floor = 4
    floor_ceiling = 3
    windows = [
        [1., 1., 1., 1.],
    ]
    thickness = 0.1
    aroom = room.create_south_facing_room(
        width=width,
        depth=depth,
        floor_floor=floor_floor,
        floor_ceiling=floor_ceiling,
        wpd=windows,
        swall_thickness=thickness,
    )



import json
import os
import tempfile
from dataclasses import asdict, dataclass, field
from pathlib import Path
from typing import List, Optional, Tuple, Union

import pyradiance as pr
import pywincalc as pwc

AIR = pwc.PredefinedGasType.AIR
KRYPTON = pwc.PredefinedGasType.KRYPTON
XENON = pwc.PredefinedGasType.XENON
ARGON = pwc.PredefinedGasType.ARGON


@dataclass
class PaneRGB:
    """Pane color data object.

    Attributes:
        measured_data: measured data as a PaneProperty object.
        coated_rgb: Coated side RGB.
        glass_rgb: Non-coated side RGB.
        trans_rgb: Transmittance RGB.
    """

    coated_rgb: Tuple[float, float, float]
    glass_rgb: Tuple[float, float, float]
    trans_rgb: Tuple[float, float, float]
    coated_side: Optional[str] = None


@dataclass
class Layer:
    """Layer data object.

    Attributes:
        product_name: Name of the product.
        thickness: Thickness of the layer.
        product_type: Type of product.
        conductivity: Conductivity of the layer.
        emissivity_front: Front emissivity of the layer.
        emissivity_back: Back emissivity of the layer.
        ir_transmittance: IR transmittance of the layer.
        rgb: PaneRGB object.
    """

    product_name: str
    thickness: float
    product_type: str
    conductivity: float
    emissivity_front: float
    emissivity_back: float
    ir_transmittance: float
    rgb: PaneRGB


@dataclass
class Gas:
    """Gas data object.

    Attributes:
        gas: Gas type.
        ratio: Gas ratio.
    """

    gas: str
    ratio: float

    def __post_init__(self):
        if self.ratio < 0 or self.ratio > 1:
            raise ValueError("Gas ratio must be between 0 and 1.")
        if self.gas.lower() not in ("air", "argon", "krypton", "xenon"):
            raise ValueError("Invalid gas type.")


@dataclass
class Gap:
    """Gap data object.

    Attributes:
        gas: List of Gas objects.
        thickness: Thickness of the gap.
    """

    gas: List[Gas]
    thickness: float

    def __post_init__(self):
        if self.thickness <= 0:
            raise ValueError("Gap thickness must be greater than 0.")
        if sum(g.ratio for g in self.gas) != 1:
            raise ValueError("The sum of the gas ratios must be 1.")


@dataclass
class GlazingSystem:
    """Glazing system data object.

    Attributes:
        name: Name of the glazing system.
        thickness: Thickness of the glazing system.
        layers: List of Layer objects.
        gaps: List of Gap objects.
        visible_front_transmittance: Visible front transmittance matrix.
        visible_back_transmittance: Visible back transmittance matrix.
        visible_front_reflectance: Visible front reflectance matrix.
        visible_back_reflectance: Visible back reflectance matrix.
        solar_front_transmittance: Solar front transmittance matrix.
        solar_back_transmittance: Solar back transmittance matrix.
        solar_front_reflectance: Solar front reflectance matrix.
        solar_back_reflectance: Solar back reflectance matrix.
        solar_front_absorptance: Solar front absorptance matrix by layer.
        solar_back_absorptance: Solar back absorptance matrix by layer.
    """

    name: str
    thickness: float = 0
    layers: List[Layer] = field(default_factory=list)
    gaps: List[Gap] = field(default_factory=list)
    visible_front_transmittance: List[List[float]] = field(default_factory=list)
    visible_back_transmittance: List[List[float]] = field(default_factory=list)
    visible_front_reflectance: List[List[float]] = field(default_factory=list)
    visible_back_reflectance: List[List[float]] = field(default_factory=list)
    solar_front_transmittance: List[List[float]] = field(default_factory=list)
    solar_back_transmittance: List[List[float]] = field(default_factory=list)
    solar_front_reflectance: List[List[float]] = field(default_factory=list)
    solar_back_reflectance: List[List[float]] = field(default_factory=list)
    solar_front_absorptance: List[List[float]] = field(default_factory=list)
    solar_back_absorptance: List[List[float]] = field(default_factory=list)

    def _matrix_to_str(self, matrix: List[List[float]]) -> str:
        """Convert a matrix to a string."""
        return "\n".join([" ".join([str(n) for n in row]) for row in matrix])

    def to_xml(self, out):
        """Save the glazing system to a file."""
        with tempfile.TemporaryDirectory() as tmpdir:
            _tmpdir = Path(tmpdir)
            with open(_tvf := _tmpdir / "tbv", "w") as f:
                f.write(self._matrix_to_str(self.visible_front_transmittance))
            with open(_tvb := _tmpdir / "tfv", "w") as f:
                f.write(self._matrix_to_str(self.visible_back_transmittance))
            with open(_rvf := _tmpdir / "rfv", "w") as f:
                f.write(self._matrix_to_str(self.visible_front_reflectance))
            with open(_rvb := _tmpdir / "rbv", "w") as f:
                f.write(self._matrix_to_str(self.visible_back_reflectance))
            with open(_tsf := _tmpdir / "tbs", "w") as f:
                f.write(self._matrix_to_str(self.solar_front_transmittance))
            with open(_tsb := _tmpdir / "tfs", "w") as f:
                f.write(self._matrix_to_str(self.solar_back_transmittance))
            with open(_rsf := _tmpdir / "rfs", "w") as f:
                f.write(self._matrix_to_str(self.solar_front_reflectance))
            with open(_rsb := _tmpdir / "rbs", "w") as f:
                f.write(self._matrix_to_str(self.solar_back_reflectance))
            _vi = pr.WrapBSDFInput("Visible", _tvf, _tvb, _rvf, _rvb)
            _si = pr.WrapBSDFInput("Solar", _tsf, _tsb, _rsf, _rsb)
            with open(out, "wb") as f:
                f.write(
                    pr.wrapbsdf(
                        enforce_window=True,
                        basis="kf",
                        inp=[_vi, _si],
                        n=self.name,
                        m="",
                        t=str(self.thickness),
                    )
                )

    def save(self, out: Union[str, Path]):
        """Save the glazing system to a file."""
        out = Path(out)
        if out.suffix == ".xml":
            self.to_xml(out)
        else:
            with open(out.with_suffix(".json"), "w") as f:
                json.dump(asdict(self), f)

    @classmethod
    def from_json(cls, path: Union[str, Path]):
        """Load a glazing system from a JSON file."""
        with open(path, "r") as f:
            data = json.load(f)
        layers = data.pop("layers")
        layer_instances = []
        for layer in layers:
            rgb = layer.pop("rgb")
            layer_instances.append(Layer(rgb=PaneRGB(**rgb), **layer))
        gaps = data.pop("gaps")
        gap_instances = []
        for gap in gaps:
            gas_instances = []
            gases = gap.pop("gas")
            for gs in gases:
                gas_instances.append(Gas(**gs))
            gap_instances.append(Gap(gas=gas_instances, **gap))
        return cls(layers=layer_instances, gaps=gap_instances, **data)

    def get_brtdfunc(self) -> pr.Primitive:
        """Get a BRTDfunc primitive for the glazing system."""
        if any(layer.product_type != "glazing" for layer in self.layers):
            raise ValueError("Only glass layers supported.")
        if len(self.layers) > 2:
            raise ValueError("Only double pane supported.")
        rgb = [layer.rgb for layer in self.layers]
        return get_glazing_primitive(self.name, rgb)


def load_glazing_system(path: Union[str, Path]):
    """Load a glazing system from a JSON file."""
    with open(path, "r") as f:
        data = json.load(f)
    layers = data.pop("layers")
    layer_instances = []
    for layer in layers:
        rgb = layer.pop("rgb")
        layer_instances.append(Layer(rgb=PaneRGB(**rgb), **layer))
    gaps = data.pop("gaps")
    gap_instances = []
    for gap in gaps:
        gas_instances = []
        gases = gap.pop("gas")
        for gs in gases:
            gas_instances.append(Gas(**gs))
        gap_instances.append(Gap(gas=gas_instances, **gap))
    return GlazingSystem(layers=layer_instances, gaps=gap_instances, **data)


def get_layers(input: List[pwc.ProductData]) -> List[Layer]:
    """Create a list of layers from a list of pwc.ProductData."""
    layers = []
    for inp in input:
        layers.append(
            Layer(
                product_name=inp.product_name,
                thickness=inp.thickness,
                product_type=inp.product_type,
                conductivity=inp.conductivity,
                emissivity_front=inp.emissivity_front,
                emissivity_back=inp.emissivity_back,
                ir_transmittance=inp.ir_transmittance,
                rgb=get_layer_rgb(inp),
            )
        )
    return layers


def create_pwc_gaps(gaps: List[Gap]):
    """Create a list of pwc gaps from a list of gaps."""
    pwc_gaps = []
    for gap in gaps:
        _gas = pwc.create_gas(
            [[g.ratio, getattr(pwc.PredefinedGasType, g.gas.upper())] for g in gap.gas]
        )
        _gap = pwc.Layers.gap(gas=_gas, thickness=gap.thickness)
        pwc_gaps.append(_gap)
    return pwc_gaps


def create_glazing_system(
    name: str,
    layers: List[Union[Path, str, bytes]],
    gaps: Optional[List[Gap]] = None,
) -> GlazingSystem:
    """Create a glazing system from a list of layers and gaps.

    Args:
        name: Name of the glazing system.
        layers: List of layers.
        gaps: List of gaps.

    Returns:
        GlazingSystem object.

    Raises:
        ValueError: Invalid layer type.

    Examples:
        >>> create_glazing_system(
        ...     "test",
        ...     [
        ...         Path("glass.json"),
        ...         Path("venetian.xml"),
        ...     ],
        ... )
    """
    if gaps is None:
        gaps = [Gap([Gas("air", 1)], 0.0127) for _ in range(len(layers) - 1)]
    layer_data = []
    thickness = 0
    for layer in layers:
        product_data = None
        if isinstance(layer, str) or isinstance(layer, Path):
            if not os.path.isfile(layer):
                raise FileNotFoundError(f"{layer} does not exist.")
            if isinstance(layer, Path):
                layer = str(layer)
            if layer.endswith(".json"):
                product_data = pwc.parse_json_file(layer)
            elif layer.endswith(".xml"):
                product_data = pwc.parse_bsdf_xml_file(layer)
                if product_data.product_type is None:
                    product_data.product_type = "shading"
            else:
                product_data = pwc.parse_optics_file(layer)
        elif isinstance(layer, bytes):
            try:
                product_data = pwc.parse_json(layer)
            except json.JSONDecodeError:
                product_data = pwc.parse_bsdf_xml_string(layer)
        if product_data is None:
            raise ValueError("Invalid layer type")
        layer_data.append(product_data)
        product_data.thickness = product_data.thickness / 1000.0 or 0  # mm to m
        thickness += product_data.thickness

    for gap in gaps:
        thickness += gap.thickness

    glzsys = pwc.GlazingSystem(
        solid_layers=layer_data,
        gap_layers=create_pwc_gaps(gaps),
        width_meters=1,
        height_meters=1,
        environment=pwc.nfrc_shgc_environments(),
        bsdf_hemisphere=pwc.BSDFHemisphere.create(pwc.BSDFBasisType.FULL),
    )

    solres = glzsys.optical_method_results("SOLAR")
    solsys = solres.system_results
    visres = glzsys.optical_method_results("PHOTOPIC")
    vissys = visres.system_results

    return GlazingSystem(
        name=name,
        thickness=thickness,
        layers=get_layers(layer_data),
        gaps=gaps,
        solar_front_absorptance=[
            alpha.front.absorptance.angular_total for alpha in solres.layer_results
        ],
        solar_back_absorptance=[
            alpha.back.absorptance.angular_total for alpha in solres.layer_results
        ],
        visible_back_reflectance=vissys.back.reflectance.matrix,
        visible_front_reflectance=vissys.front.reflectance.matrix,
        visible_back_transmittance=vissys.back.transmittance.matrix,
        visible_front_transmittance=vissys.front.transmittance.matrix,
        solar_back_reflectance=solsys.back.reflectance.matrix,
        solar_front_reflectance=solsys.front.reflectance.matrix,
        solar_back_transmittance=solsys.back.transmittance.matrix,
        solar_front_transmittance=solsys.front.transmittance.matrix,
    )


def get_layer_rgb(layer: pwc.ProductData) -> PaneRGB:
    """Get the RGB values for a pane layer."""
    photopic_wvl = range(380, 781, 10)
    if isinstance(layer.measurements, pwc.DualBandBSDF):
        return PaneRGB(
            coated_rgb=(0, 0, 0),
            glass_rgb=(0, 0, 0),
            trans_rgb=(0, 0, 0),
            coated_side=None,
        )
    hemi = {
        int(round(d.wavelength * 1e3)): (
            d.direct_component.transmittance_front,
            d.direct_component.transmittance_back,
            d.direct_component.reflectance_front,
            d.direct_component.reflectance_back,
        )
        for d in layer.measurements
    }
    tvf = [hemi[w][0] for w in photopic_wvl]
    rvf = [hemi[w][2] for w in photopic_wvl]
    rvb = [hemi[w][3] for w in photopic_wvl]
    tf_x, tf_y, tf_z = pr.spec_xyz(tvf, 380, 780)
    rf_x, rf_y, rf_z = pr.spec_xyz(rvf, 380, 780)
    rb_x, rb_y, rb_z = pr.spec_xyz(rvb, 380, 780)
    tf_rgb = pr.xyz_rgb(tf_x, tf_y, tf_z)
    rf_rgb = pr.xyz_rgb(rf_x, rf_y, rf_z)
    rb_rgb = pr.xyz_rgb(rb_x, rb_y, rb_z)
    if layer.coated_side == "front":
        coated_rgb = rf_rgb
        glass_rgb = rb_rgb
    else:
        coated_rgb = rb_rgb
        glass_rgb = rf_rgb
    return PaneRGB(coated_rgb, glass_rgb, tf_rgb, layer.coated_side)


def get_glazing_primitive(name: str, panes: List[PaneRGB]) -> pr.Primitive:
    """Generate a BRTDfunc to represent a glazing system."""
    if len(panes) > 2:
        raise ValueError("Only double pane supported")
    if len(panes) == 1:
        str_arg = [
            "sr_clear_r",
            "sr_clear_g",
            "sr_clear_b",
            "st_clear_r",
            "st_clear_g",
            "st_clear_b",
            "0",
            "0",
            "0",
            "glaze1.cal",
        ]
        coated_real = 1 if panes[0].coated_side == "front" else -1
        real_arg = [
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            coated_real,
            *[round(i, 3) for i in panes[0].glass_rgb],
            *[round(i, 3) for i in panes[0].coated_rgb],
            *[round(i, 3) for i in panes[0].trans_rgb],
        ]
    else:
        s12t_r, s12t_g, s12t_b = panes[0].trans_rgb
        s34t_r, s34t_g, s34t_b = panes[1].trans_rgb
        if panes[0].coated_side == "back":
            s2r_r, s2r_g, s2r_b = panes[0].coated_rgb
            s1r_r, s1r_g, s1r_b = panes[0].glass_rgb
        else:  # front or neither side coated
            s2r_r, s2r_g, s2r_b = panes[0].glass_rgb
            s1r_r, s1r_g, s1r_b = panes[0].coated_rgb
        if panes[1].coated_side == "back":
            s4r_r, s4r_g, s4r_b = panes[1].coated_rgb
            s3r_r, s3r_g, s3r_b = panes[1].glass_rgb
        else:  # front or neither side coated
            s4r_r, s4r_g, s4r_b = panes[1].glass_rgb
            s3r_r, s3r_g, s3r_b = panes[1].coated_rgb
        str_arg = [
            (
                f"if(Rdot,cr(fr({s4r_r:.3f}),ft({s34t_r:.3f}),fr({s2r_r:.3f})),"
                f"cr(fr({s1r_r:.3f}),ft({s12t_r:.3f}),fr({s3r_r:.3f})))"
            ),
            (
                f"if(Rdot,cr(fr({s4r_g:.3f}),ft({s34t_g:.3f}),fr({s2r_g:.3f})),"
                f"cr(fr({s1r_g:.3f}),ft({s12t_g:.3f}),fr({s3r_g:.3f})))"
            ),
            (
                f"if(Rdot,cr(fr({s4r_b:.3f}),ft({s34t_b:.3f}),fr({s2r_b:.3f})),"
                f"cr(fr({s1r_b:.3f}),ft({s12t_b:.3f}),fr({s3r_b:.3f})))"
            ),
            f"ft({s34t_r:.3f})*ft({s12t_r:.3f})",
            f"ft({s34t_g:.3f})*ft({s12t_g:.3f})",
            f"ft({s34t_b:.3f})*ft({s12t_b:.3f})",
            "0",
            "0",
            "0",
            "glaze2.cal",
        ]
        real_arg = [0, 0, 0, 0, 0, 0, 0, 0, 0]
    return pr.Primitive("void", "BRTDfunc", name, str_arg, real_arg)



"""
This module contains all utility functions used throughout frads.
"""
from datetime import datetime, timedelta
from io import TextIOWrapper
import logging
import re
from pathlib import Path
from random import choices
import string
import subprocess as sp
from typing import Any, Dict, Optional, List, Union

from frads import geom
import numpy as np
from pyradiance import Primitive, parse_primitive, pvaluer


logger: logging.Logger = logging.getLogger("frads.utils")


def parse_rad_header(header_str: str) -> tuple:
    """Parse a Radiance matrix file header.

    Args:
        header_str: header as string
    Returns:
        A tuple contain nrow, ncol, ncomp, datatype
    Raises:
        ValueError: if any of NROWS NCOLS NCOMP FORMAT is not found.
    """
    compiled = re.compile(
        r" NROWS=(.*) | NCOLS=(.*) | NCOMP=(.*) | FORMAT=(.*) ", flags=re.X
    )
    matches = compiled.findall(header_str)
    if len(matches) != 4:
        raise ValueError("Can't find one of the header entries.")
    nrow = int([mat[0] for mat in matches if mat[0] != ""][0])
    ncol = int([mat[1] for mat in matches if mat[1] != ""][0])
    ncomp = int([mat[2] for mat in matches if mat[2] != ""][0])
    dtype = [mat[3] for mat in matches if mat[3] != ""][0].strip()
    return nrow, ncol, ncomp, dtype


def polygon_primitive(
    polygon: geom.Polygon, modifier: str, identifier: str
) -> Primitive:
    """
    Generate a primitive from a polygon.
    Args:
        polygon: a Polygon object
        modifier: a Radiance primitive modifier
        identifier: a Radiance primitive identifier
    Returns:
        A Primitive object
    """
    return Primitive(modifier, "polygon", identifier, [], polygon.coordinates)


def parse_polygon(primitive: Primitive) -> geom.Polygon:
    """
    Parse a primitive into a polygon.

    Args:
        primitive: a dictionary object containing a primitive
    Returns:
        A Polygon object
    """
    if primitive.ptype != "polygon":
        raise ValueError("Not a polygon: ", primitive.identifier)
    vertices = [
        np.array(primitive.fargs[i : i + 3]) for i in range(0, len(primitive.fargs), 3)
    ]
    return geom.Polygon(vertices)


def array_hdr(array: np.ndarray, xres: int, yres: int, dtype: str = "d") -> bytes:
    """
    Call pvalue to generate a HDR image from a numpy array.
    Args:
        array: one-dimensional pixel values [[r1, g1, b1], [r2, g2, b2], ...]
        xres: x resolution
        yres: y resolution
        dtype: data type of the array. 'd' for double, 'f' for float
    Returns:
        HDR image in bytes
    """
    return pvaluer(array.tobytes(), inform=dtype, header=False, xres=xres, yres=yres)


def write_hdr(
    fname: str, array: np.ndarray, xres: int, yres: int, dtype: str = "d"
) -> None:
    """
    Write a array into a HDR image.
    Args:
        fname: output file name
        array: one-dimensional pixel values [[r1, g1, b1], [r2, g2, b2], ...]
        xres: x resolution
        yres: y resolution
        dtype: data type of the array. 'd' for double, 'f' for float
    Returns:
        None
    """
    with open(fname, "wb") as f:
        f.write(array_hdr(array, xres, yres, dtype=dtype))


def write_hdrs(
    array: np.ndarray, xres: int, yres: int, dtype: str = "d", outdir: str = "."
) -> None:
    """
    Write a series of HDR images to a file.
    Args:
        array: two-dimensional pixel values [[r1, g1, b1], [r2, g2, b2], ...]
            where each column of data represents a image.
        xres: x resolution
        yres: y resolution
        dtype: data type of the array. 'd' for double, 'f' for float
        outdir: output directory
    Returns:
        None
    """
    if not Path(outdir).exists():
        Path(outdir).mkdir(parents=True)
    # iterate through the columns of array
    for i in range(array.shape[1]):
        fname = f"{outdir}/hdr_{i}.hdr"
        write_hdr(fname, array[:, i], xres, yres, dtype=dtype)


def write_ep_rad_model(outpath: str, model: dict) -> None:
    """
    Write a epjson2rad model into a Radiance file.
    Args:
        outpath: output file path
        model: a model object
    """
    with open(outpath, "wb") as f:
        f.write(model["model"]["materials"]["bytes"])
        f.write(model["model"]["scene"]["bytes"])
        for window in model["model"]["windows"].values():
            f.write(window["bytes"])


def unpack_primitives(file: Union[str, Path, TextIOWrapper]) -> List[Primitive]:
    """Open a file a to parse primitive."""
    if isinstance(file, TextIOWrapper):
        lines = file.read()
    else:
        with open(file, "r", encoding="ascii") as rdr:
            lines = rdr.read()
    return parse_primitive(lines)


def primitive_normal(primitive_paths: List[str]) -> List[np.ndarray]:
    """Return a set of normal vectors given a list of primitive paths."""
    _primitives: List[Primitive] = []
    for path in primitive_paths:
        _primitives.extend(unpack_primitives(path))
    seen = {}
    unique = []
    _normals = [parse_polygon(prim).normal for prim in _primitives]
    for n in _normals:
        nstr = n.tobytes()
        if nstr not in seen:
            seen[nstr] = True
            unique.append(n)
    return unique


def neutral_plastic_prim(
    mod: str, ident: str, refl: float, spec: float, rough: float
) -> Primitive:
    """
    Generate a neutral color plastic material.

    Args:
        mod: modifier to the primitive
        ident: identifier to the primitive
        refl: measured reflectance (0.0 - 1.0)
        spec: material specularity (0.0 - 1.0)
        rough: material roughness (0.0 - 1.0)

    Returns:
        A material primtive
    """
    err_msg = "reflectance, speculariy, and roughness have to be 0-1"
    assert all(0 <= i <= 1 for i in [spec, refl, rough]), err_msg
    real_args = [refl, refl, refl, spec, rough]
    return Primitive(mod, "plastic", ident, [], real_args)


def neutral_trans_prim(
    mod: str, ident: str, trans: float, refl: float, spec: float, rough: float
) -> Primitive:
    """
    Generate a neutral color plastic material.

    Args:
        mod: modifier to the primitive
        ident: identifier to the primitive
        refl: measured reflectance (0.0 - 1.0)
        spec: material specularity (0.0 - 1.0)
        rough: material roughness (0.0 - 1.0)

    Returns:
        A material primtive
    """
    color = trans + refl
    t_diff = trans / color
    tspec = 0
    err_msg = "reflectance, speculariy, and roughness have to be 0-1"
    assert all(0 <= i <= 1 for i in [spec, refl, rough]), err_msg
    real_args = [color, color, color, spec, rough, t_diff, tspec]
    return Primitive(mod, "trans", ident, [], real_args)


def color_plastic_prim(
    mod: str,
    ident: str,
    refl: float,
    red: int,
    green: int,
    blue: int,
    specu: float,
    rough: float,
) -> Primitive:
    """Generate a colored plastic material.
    Args:
        mod: modifier to the primitive
        ident: identifier to the primitive
        refl : measured reflectance (0.0 - 1.0)
        red: green; blue (int): rgb values (0 - 255)
        specu: material specularity (0.0 - 1.0)
        rough: material roughness (0.0 - 1.0)

    Returns:
        A material primtive
    """
    err_msg = "reflectance, speculariy, and roughness have to be 0-1"
    assert all(0 <= i <= 1 for i in [specu, refl, rough]), err_msg
    red_eff = 0.3
    green_eff = 0.59
    blue_eff = 0.11
    weighted = red * red_eff + green * green_eff + blue * blue_eff
    matr = round(red / weighted * refl, 3)
    matg = round(green / weighted * refl, 3)
    matb = round(blue / weighted * refl, 3)
    real_args = [matr, matg, matb, specu, rough]
    return Primitive(mod, "plastic", ident, [], real_args)


def add_manikin(
    manikin_file: str,
    manikin_name: str,
    zone: dict,
    position: List[float],
    rotation: float = 0,
) -> None:
    """Add a manikin to the scene.i
    Args:
        manikin_file: path to the manikin file
        manikin_name: name of the manikin
        zone: zone as a dictionary, must have 'model' key, we assume all scene
            data is inside the data key, and not files.
        position: position of the manikin (x, y), where x and y are 0-1
        rotation: rotation of the manikin in degree (0-360)
    Returns:
        A zone with added manikin
    Notes:
        Zone dictionary is modified in place.
    """
    zone["model"]["scene"]["bytes"] += b" "
    zone_primitives = parse_primitive(zone["model"]["scene"]["bytes"].decode())
    zone_polygons = [parse_polygon(p) for p in zone_primitives if p.ptype == "polygon"]
    xmin, xmax, ymin, ymax, zmin, _ = geom.get_polygon_limits(zone_polygons)
    target = np.array(
        [xmin + (xmax - xmin) * position[0], ymin + (ymax - ymin) * position[1], zmin]
    )
    with open(manikin_file) as f:
        manikin_primitives = parse_primitive(f.read())
    non_polygon_primitives = [p for p in manikin_primitives if p.ptype != "polygon"]
    for primitive in non_polygon_primitives:
        zone["model"]["scene"]["bytes"] += primitive.bytes
    manikin_polygons = [
        parse_polygon(p) for p in manikin_primitives if p.ptype == "polygon"
    ]
    xminm, xmaxm, yminm, ymaxm, zminm, _ = geom.get_polygon_limits(manikin_polygons)
    manikin_base_center = np.array([(xmaxm - xminm) / 2, (ymaxm - yminm) / 2, zminm])
    if rotation != 0:
        manikin_polygons = [
            p.rotate(manikin_base_center, np.array([0, 0, 1]), np.radians(rotation))
            for p in manikin_polygons
        ]
    move_vector = manikin_base_center - target
    moved_manikin_polygons = [polygon.move(move_vector) for polygon in manikin_polygons]
    moved_manikin = [
        polygon2prim(polygon, primitive.modifier, primitive.identifier)
        for polygon, primitive in zip(moved_manikin_polygons, manikin_primitives)
    ]
    for primitive in moved_manikin:
        zone["model"]["scene"]["bytes"] += primitive.bytes
    manikin_rays = []
    for polygon in moved_manikin_polygons:
        manikin_rays.append([*polygon.centroid.tolist(), *polygon.normal.tolist()])
    zone["model"]["sensors"][manikin_name] = {"data": manikin_rays}


def glass_prim(
    mod: str, ident: str, tr: float, tg: float, tb: float, refrac: float = 1.52
) -> Primitive:
    """Generate a glass material.

    Args:
        mod: modifier to the primitive
        ident: identifier to the primtive
        tr: Transmissivity in red channel (0.0 - 1.0)
        tg: Transmissivity in green channel (0.0 - 1.0)
        tb: Transmissivity in blue channel (0.0 - 1.0)
        refrac: refraction index (default=1.52)
    Returns:
        material primtive (dict)

    """
    tmsv_red = tr * 1.08981
    tmsv_green = tg * 1.08981
    tmsv_blue = tb * 1.08981
    real_args = [tmsv_red, tmsv_green, tmsv_blue, refrac]
    return Primitive(mod, "glass", ident, [], real_args)


def bsdf_prim(
    mod: str,
    ident: str,
    xmlpath: str,
    upvec: List[float],
    pe: bool = False,
    thickness: float = 0.0,
    xform=None,
    real_args: list = [""],
) -> Primitive:
    """Create a BSDF primtive."""
    str_args = [xmlpath, str(upvec)]
    str_args_count = 5
    if pe:
        _type = "aBSDF"
    else:
        str_args_count += 1
        str_args = [str(thickness), *str_args]
        _type = "BSDF"
    if xform is not None:
        str_args_count += len(xform.split())
        str_args.extend(*xform.split())
    else:
        str_args.append(".")
    return Primitive(mod, _type, ident, str_args, real_args)


def opt2list(opt: dict) -> List[str]:
    """Convert option dictionary to list.

    Key: str
    Value: str | float | int | bool | list

    Args:
        opt: option dictionary
    Returns:
        A list of strings
    """
    out = []
    for key, value in opt.items():
        if isinstance(value, str):
            if key == "vf":
                out.extend(["-" + key, value])
            else:
                out.append(f"-{key}{value}")
        elif isinstance(value, bool):
            if value:
                out.append(f"-{key}+")
            else:
                out.append(f"-{key}-")
        elif isinstance(value, (int, float)):
            out.extend(["-" + key, str(value)])
        elif isinstance(value, list):
            out.extend(["-" + key, *map(str, value)])
    return out


def pt_inclusion(pt: np.ndarray, polygon_pts: List[np.ndarray]) -> int:
    """Test whether a point is inside a polygon
    using winding number algorithm."""

    def isLeft(pt0, pt1, pt2):
        """Test whether a point is left to a line."""
        return (pt1[0] - pt0[0]) * (pt2[1] - pt0[1]) - (pt2[0] - pt0[0]) * (
            pt1[1] - pt0[1]
        )

    # Close the polygon for looping
    # polygon_pts.append(polygon_pts[0])
    polygon_pts = [*polygon_pts, polygon_pts[0]]
    wn = 0
    for i in range(len(polygon_pts) - 1):
        if polygon_pts[i][1] <= pt[1]:
            if polygon_pts[i + 1][1] > pt[1]:
                if isLeft(polygon_pts[i], polygon_pts[i + 1], pt) > 0:
                    wn += 1
        else:
            if polygon_pts[i + 1][1] <= pt[1]:
                if isLeft(polygon_pts[i], polygon_pts[i + 1], pt) < 0:
                    wn -= 1
    return wn


def gen_grid(polygon: geom.Polygon, height: float, spacing: float) -> List[List[float]]:
    """Generate a grid of points for orthogonal planar surfaces.

    Args:
        polygon: a polygon object
        height: points' distance from the surface in its normal direction
        spacing: distance between the grid points
    Returns:
        List of the points as list
    """
    vertices = polygon.vertices
    plane_height = sum(i[2] for i in vertices) / len(vertices)
    imin, imax, jmin, jmax, _, _ = polygon.extreme
    xlen_spc = (imax - imin) / spacing
    ylen_spc = (jmax - jmin) / spacing
    xstart = ((xlen_spc - int(xlen_spc) + 1)) * spacing / 2
    ystart = ((ylen_spc - int(ylen_spc) + 1)) * spacing / 2
    x0 = np.arange(imin, imax, spacing) + xstart
    y0 = np.arange(jmin, jmax, spacing) + ystart
    grid_dir = polygon.normal * -1
    grid_hgt = np.array((0, 0, plane_height)) + grid_dir * height
    raw_pts = [
        np.array((round(i, 3), round(j, 3), round(grid_hgt[2], 3)))
        for i in x0
        for j in y0
    ]
    if np.array_equal(polygon.normal, np.array((0, 0, 1))):
        _grid = [p for p in raw_pts if pt_inclusion(p, vertices) > 0]
    else:
        _grid = [p for p in raw_pts if pt_inclusion(p, vertices[::-1]) > 0]
    grid = [p.tolist() + grid_dir.tolist() for p in _grid]
    return grid


def material_lib() -> Dict[str, Primitive]:
    """Generate a list of generic material primitives."""
    tmis = 0.6 * 1.08981
    return {
        "neutral_lambertian_0.2": neutral_plastic_prim(
            "void", "neutral_lambertian_0.2", 0.2, 0, 0
        ),
        "neutral_lambertian_0.5": neutral_plastic_prim(
            "void", "neutral_lambertian_0.5", 0.5, 0, 0
        ),
        "neutral_lambertian_0.7": neutral_plastic_prim(
            "void", "neutral_lambertian_0.7", 0.7, 0, 0
        ),
        "glass_60": glass_prim("void", "glass_60", tmis, tmis, tmis),
    }


def gen_blinds(depth, width, height, spacing, angle, curve, movedown) -> str:
    """Generate genblinds command for genBSDF."""
    nslats = int(round(height / spacing, 0))
    slat_cmd = "!genblinds blindmaterial blinds "
    slat_cmd += f"{depth} {width} {height} {nslats} {angle} {curve}"
    slat_cmd += "| xform -rz -90 -rx -90 -t "
    slat_cmd += f"{-width/2} {-height/2} {-movedown}\n"
    return slat_cmd


def batch_process(
    commands: List[List[str]],
    inputs: Optional[List[bytes]] = None,
    opaths: Optional[List[Path]] = None,
    nproc: Optional[int] = None,
) -> None:
    """Run commands in batches.

    Use subprocess.Popen to run commands.

    Args:
        commands: commands as a list of strings.
        inputs: list of standard input to the commands.
        opaths: list of paths to write standard output to.
        nproc: number of commands to run in parallel at a time.
    Returns:
        None
    """
    nproc = 1 if nproc is None else nproc
    command_groups = [commands[i : i + nproc] for i in range(0, len(commands), nproc)]
    stdin_groups: List[List[Any]] = [[None] * len(ele) for ele in command_groups]
    if inputs:
        if len(inputs) != len(commands):
            raise ValueError("Number of stdins not equal number of commands.")
        stdin_groups = [inputs[i : i + nproc] for i in range(0, len(inputs), nproc)]
    if opaths:
        if len(opaths) != len(commands):
            raise ValueError("Number of opaths not equal number of commands.")
        opath_groups = [opaths[i : i + nproc] for i in range(0, len(opaths), nproc)]
        for igrp, cgrp, ogrp in zip(stdin_groups, command_groups, opath_groups):
            processes = []
            fds = []
            for command, opath in zip(cgrp, ogrp):
                fd = open(opath, "wb")
                processes.append(sp.Popen(command, stdin=sp.PIPE, stdout=fd))
                fds.append(fd)
            for proc, sin in zip(processes, igrp):
                if (proc.stdin is not None) and (sin is not None):
                    proc.stdin.write(sin)
            for p, f in zip(processes, fds):
                p.wait()
                f.close()
    else:
        for igrp, cgrp in zip(stdin_groups, command_groups):
            processes = [sp.Popen(command, stdin=sp.PIPE) for command in cgrp]
            for proc, sin in zip(processes, igrp):
                if (proc.stdin is not None) and (sin is not None):
                    proc.stdin.write(sin)
            for proc in processes:
                proc.wait()


def random_string(size: int) -> str:
    """Generate random characters."""
    chars = string.ascii_uppercase + string.digits
    return "".join(choices(chars, k=size))


def minutes_to_datetime(year: int, minutes: int):
    return datetime(year, 1, 1) + timedelta(minutes=minutes)



"""
Routines for generating sky models
"""

import datetime
import logging
import math
import os
from pathlib import Path
from typing import List, NamedTuple, Optional, Sequence, Tuple, Union

import pyradiance as pr

logger: logging.Logger = logging.getLogger("frads.sky")


class WeaMetaData(NamedTuple):
    """Weather related meta data object.

    Attributes:
        city: City.
        country: Country.
        latitude: Latitude.
        longitude: Longitude.
        timezone: Timezone as standard meridian.
        elevation: Site elevation (m).
    """

    city: str
    country: str
    latitude: float
    longitude: float
    timezone: int
    elevation: float

    def wea_header(self) -> str:
        """Return a .wea format header."""
        return (
            f"place {self.city}_{self.country}\n"
            f"latitude {self.latitude}\n"
            f"longitude {self.longitude}\n"
            f"time_zone {self.timezone}\n"
            f"site_elevation {self.elevation}\n"
            "weather_data_file_units 1\n"
        )


class WeaData(NamedTuple):
    """Weather related data object.

    Attributes:
        month: Month.
        day: Day.
        hour: Hour.
        minute: Minutes.
        second: Seconds.
        hours: Times with minutes as fraction.
        dni: Direct normal irradiance (W/m2).
        dhi: Diffuse horizontal irradiance (W/m2).
        aod: Aeroal Optical Depth (default = 0).
        cc: Cloud cover (default = 0).
        year: default = 2000.
    """

    time: datetime.datetime
    dni: float
    dhi: float
    aod: float = 0
    cc: float = 0

    def __str__(self) -> str:
        return f"{self.time.month} {self.time.day} {self.time.hour+self.time.minute/60} {self.dni} {self.dhi}"

    def dt_str(self) -> str:
        return f"{self.time.month:02d}{self.time.day:02d}_{self.time.hour:02d}{self.time.minute:02d}"


def parse_epw(epw_str: str) -> tuple:
    """Parse epw file and return wea header and data.

    Args:
        epw_str: String containing epw file.

    Returns:
        Tuple of meta data and wea data.
    """
    raw = epw_str.splitlines()
    epw_header = raw[0].split(",")
    content = raw[8:]
    data = []
    for li in content:
        line = li.split(",")
        year = int(line[0])
        month = int(line[1])
        day = int(line[2])
        hour = int(line[3]) - 1
        dir_norm = float(line[14])
        dif_hor = float(line[15])
        cc = float(line[19])
        aod = float(line[26])
        data.append(
            WeaData(
                datetime.datetime(year, month, day, hour, 30),
                dir_norm,
                dif_hor,
                cc,
                aod,
            )
        )
    city = epw_header[1]
    country = epw_header[3]
    latitude = float(epw_header[6])
    longitude = -1 * float(epw_header[7])
    tz = int(float(epw_header[8])) * (-15)
    elevation = float(epw_header[9].rstrip())
    meta_data = WeaMetaData(city, country, latitude, longitude, tz, elevation)
    return meta_data, data


def parse_wea(wea_str: str) -> Tuple[WeaMetaData, List[WeaData]]:
    """
    Parse a wea file in its entirety.
    Args:
        wea_str: String containing wea file.
    Returns:
        Tuple of meta data and wea data.
    """
    lines = wea_str.splitlines()
    place = lines[0].split(" ", 1)[1]
    lat = float(lines[1].split(" ", 1)[1])
    lon = float(lines[2].split(" ", 1)[1])
    tz = int(float(lines[3].split(" ", 1)[1]))
    ele = float(lines[4].split(" ", 1)[1])
    meta_data = WeaMetaData(place, "", lat, lon, tz, ele)
    year = datetime.datetime.today().year
    data = []
    for li in lines[6:]:
        if li.strip() == "":
            continue
        line = li.split()
        month = int(line[0])
        day = int(line[1])
        hours = float(line[2])
        hour = int(hours)
        minute = int((hours - hour) * 60)
        dir_norm = float(line[3])
        dif_hor = float(line[4])
        data.append(
            WeaData(
                datetime.datetime(year, month, day, hour, minute), dir_norm, dif_hor
            )
        )
    return meta_data, data


def parse_epw_file(file: os.PathLike) -> tuple:
    """Parse an epw file using parse_epw.

    Args:
        file: Path to epw file.

    Returns:
        Tuple of meta data and epw data.
    """
    with open(file, "r") as f:
        epw_str = f.read()
    return parse_epw(epw_str)


def parse_wea_file(file: os.PathLike) -> tuple:
    """Parse a wea file using parse_wea.

    Args:
        file: Path to wea file.

    Returns:
        Tuple of meta data and wea data.
    """
    with open(file, "r") as f:
        wea_str = f.read()
    return parse_wea(wea_str)


def genskymtx(
    data: Optional[Sequence[WeaData]] = None,
    meta: Optional[WeaMetaData] = None,
    wpath: Optional[Union[str, Path]] = None,
    onesun: bool = False,
    header: bool = True,
    average: bool = False,
    sun_only: bool = False,
    sky_only: bool = False,
    sun_file: Optional[str] = None,
    sun_mods: Optional[str] = None,
    daylight_hours_only: bool = False,
    sky_color: Optional[List[float]] = None,
    ground_color: Optional[List[float]] = None,
    rotate: Optional[float] = None,
    outform: Optional[str] = None,
    solar_radiance: bool = False,
    mfactor: int = 1,
) -> bytes:
    """Call gendaymtx to generate a sky/sun matrix
    Write results to out.  It takes either a .wea file path
    or wea data and metadata (defined in frads.types).
    If both are provided, .wea file path will be used.

    Args:
        data: A list of WeaData objects.
        meta: A WeaMetaData object.
        wpath: A .wea file path.
        onesun: If True, only one sun will be generated.
        header: If True, a header will be included in the output.
        average: If True, the output will be averaged.
        sun_only: If True, only sun will be generated.
        sky_only: If True, only sky will be generated.
        sun_file: A sun file path.
        sun_mods: A sun modifier.
        daylight_hours_only: If True, only daylight hours will be generated.
        sky_color: A list of sky color values.
        ground_color: A list of ground color values.
        rotate: A rotation value.
        outform: An output format.
        solar_radiance: If True, solar radiance will be generated.
        mfactor: An mfactor value.

    Returns:
        A bytes object containing the output.

    Raises:
        ValueError: An error occurs if neither a .wea path nor wea data is provided.
    """
    if wpath is None:
        if data is not None and meta is not None:
            inp = gen_wea(
                [d.time for d in data],
                [d.dni for d in data],
                [d.dhi for d in data],
                meta.latitude,
                meta.longitude,
                meta.timezone,
                elevation=meta.elevation,
                location=meta.city + meta.country,
            ).encode()
        else:
            raise ValueError("Either a .wea path or wea data is required.")
    else:
        inp = wpath
    _out = pr.gendaymtx(
        inp,
        header=header,
        average=average,
        sun_only=sun_only,
        sky_only=sky_only,
        solar_radiance=solar_radiance,
        sun_file=sun_file,
        sun_mods=sun_mods,
        daylight_hours_only=daylight_hours_only,
        sky_color=sky_color,
        ground_color=ground_color,
        rotate=rotate,
        outform=outform,
        onesun=onesun,
        mfactor=mfactor,
    )
    return _out


def gen_perez_sky(
    dt: datetime.datetime,
    latitude: float,
    longitude: float,
    timezone: int,
    year: Optional[int] = None,
    dirnorm: Optional[float] = None,
    diffhor: Optional[float] = None,
    dirhor: Optional[float] = None,
    dirnorm_illum: Optional[float] = None,
    diffhor_illum: Optional[float] = None,
    solar: bool = False,
    grefl: Optional[float] = None,
    rotate: Optional[float] = None,
) -> bytes:
    """Generate a perez sky using gendaylit.

    Args:
        dt: A datetime object.
        latitude: A latitude value.
        longitude: A longitude value.
        timezone: A timezone value.
        year: A year value.
        dirnorm: A direct normal value.
        diffhor: A diffuse horizontal value.
        dirhor: A direct horizontal value.
        dirnorm_illum: A direct normal illuminance value.
        diffhor_illum: A diffuse horizontal illuminance value.
        solar: If True, solar will be generated.
        grefl: A ground reflectance value.
        rotate: A rotation value.

    Returns:
        bytes: the sky primitive.
    """
    sun = pr.gendaylit(
        dt,
        latitude,
        longitude,
        timezone,
        year,
        dirnorm=dirnorm,
        diffhor=diffhor,
        dirhor=dirhor,
        dirnorm_illum=dirnorm_illum,
        diffhor_illum=diffhor_illum,
        solar=solar,
        grefl=grefl,
    )
    if rotate:
        sun = pr.xform(sun, rotatez=rotate)
    out = [pr.Primitive("skyfunc", "glow", "sglow", [], [1, 1, 1, 0]).bytes]
    out.append(pr.Primitive("sglow", "source", "sky", [], [0, 0, 1, 180]).bytes)
    out.append(pr.Primitive("sglow", "source", "ground", [], [0, 0, -1, 180]).bytes)
    return sun + b" ".join(out)


def gen_wea(
    datetimes: Sequence[datetime.datetime],
    dirnorm: Sequence[float],
    diffhor: Sequence[float],
    latitude: float,
    longitude: float,
    timezone: int,
    elevation: Optional[float] = None,
    location: Optional[str] = None,
) -> str:
    """Generate wea file from datetime, location, and sky."""
    if len(datetimes) != len(dirnorm) != len(diffhor):
        raise ValueError("datetimes, dirnorm, and diffhor must be the same length")
    rows = []
    if location is None or location == "":
        location = "_".join(
            [str(i) for i in [latitude, longitude, timezone, elevation]]
        )
    if elevation is None:
        elevation = 0
    rows.append(f"place {location}")
    rows.append(f"latitude {latitude}")
    rows.append(f"longitude {longitude}")
    rows.append(f"time_zone {timezone}")
    rows.append(f"site_elevation {elevation}")
    rows.append("weather_data_file_units 1")
    for dt, dni, dhi in zip(datetimes, dirnorm, diffhor):
        _hrs = dt.hour + dt.minute / 60  # middle of hour
        _row = f"{dt.month} {dt.day} {_hrs} {dni} {dhi}"
        rows.append(_row)
    return "\n".join(rows)


def solar_angle(
    lat: float,
    lon: float,
    mer: float,
    month,
    day: int,
    hour,
) -> Tuple[float, float]:
    """
    Simplified translation from the Radiance sun.c and gensky.c code.
    """
    latitude_r = math.radians(lat)
    longitude_r = math.radians(lon)
    s_meridian = math.radians(mer)
    mo_da = [0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334]

    julian_date = mo_da[month - 1] + day

    solar_decline = 0.4093 * math.sin((2 * math.pi / 365) * (julian_date - 81))

    solar_time = hour + (
        0.170 * math.sin((4 * math.pi / 373) * (julian_date - 80))
        - 0.129 * math.sin((2 * math.pi / 355) * (julian_date - 8))
        + (12 / math.pi) * (s_meridian - longitude_r)
    )
    altitude = math.asin(
        math.sin(latitude_r) * math.sin(solar_decline)
        - math.cos(latitude_r)
        * math.cos(solar_decline)
        * math.cos(solar_time * (math.pi / 12))
    )
    azimuth = -math.atan2(
        math.cos(solar_decline) * math.sin(solar_time * (math.pi / 12.0)),
        -math.cos(latitude_r) * math.sin(solar_time)
        - math.sin(latitude_r)
        * math.cos(solar_decline)
        * math.cos(solar_time * (math.pi / 12)),
    )

    return altitude, azimuth


def start_end_hour(data: Sequence[WeaData], sh: float, eh: float) -> Sequence[WeaData]:
    """Remove wea data entries outside of the
    start and end hour."""
    if sh == 0 and eh == 0:
        return data
    return [row for row in data if sh <= (row.time.hour + row.time.minute / 60) <= eh]


def check_sun_above_horizon(data, metadata):
    """Remove non-daylight hour entries."""

    def solar_altitude_check(row: WeaData):
        alt, _ = solar_angle(
            metadata.latitude,
            metadata.longitude,
            metadata.timezone,
            row.time.month,
            row.time.day,
            row.time.hour + row.time.minute / 60,
        )
        return alt > 0

    return [row for row in data if solar_altitude_check(row)]


def filter_data_with_zero_dni(data):
    """Filter out data entries with zero direct normal irradiance."""
    return [row for row in data if row.dni != 0]


def solar_minute(data: WeaData) -> int:
    # assuming never leap year
    mo_da = [0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334]
    jd = mo_da[data.time.month - 1] + data.time.day
    return 24 * 60 * (jd - 1) + int(data.time.hour * 60.0 + data.time.minute + 0.5)



"""
"""
# from frads import epjson_to_rad
from frads import epmodel_to_radmodel
from frads.eplus import load_energyplus_model


def test_cfscase(resources_dir):
    epjson1 = resources_dir / "CmplxGlz_SingleZone_DoubleClearAir.epJSON"
    epmodel = load_energyplus_model(epjson1)
    rad_model = epmodel_to_radmodel(epmodel)
    assert (
        rad_model["Room 102"]["model"]["sensors"]["Room102 Floor"] is not None
    )


def test_refcase(resources_dir):
    epjson2 = resources_dir / "RefBldgMediumOfficeNew2004_Chicago_epJSON.epJSON"
    epmodel = load_energyplus_model(epjson2)
    rad_model = epmodel_to_radmodel(epmodel)
    assert (
        rad_model["Perimeter_bot_ZN_1"]["model"]["sensors"][
            "Perimeter_bot_ZN_1_Floor"
        ]
        is not None
    )



from datetime import datetime

import pyradiance as pr
import pytest
from pyenergyplus.dataset import ref_models, weather_files

from frads.ep2rad import epmodel_to_radmodel
from frads.eplus import load_energyplus_model
from frads.methods import (
    MaterialConfig,
    Model,
    SceneConfig,
    SensorConfig,
    Settings,
    SurfaceConfig,
    ThreePhaseMethod,
    TwoPhaseMethod,
    ViewConfig,
    WindowConfig,
    WorkflowConfig,
)
from frads.window import Gap, Gas, create_glazing_system


@pytest.fixture
def cfg(resources_dir, objects_dir):
    return {
        "settings": {
            "method": "2phase",
            "sky_basis": "r1",
            "epw_file": "",
            "wea_file": resources_dir / "oak.wea",
            "sensor_sky_matrix": ["-ab", "0"],
            "view_sky_matrix": ["-ab", "0"],
            "sensor_window_matrix": ["-ab", "0"],
            "view_window_matrix": ["-ab", "0"],
            "daylight_matrix": ["-ab", "0"],
        },
        "model": {
            "scene": {
                "files": [
                    objects_dir / "walls.rad",
                    objects_dir / "ceiling.rad",
                    objects_dir / "floor.rad",
                    objects_dir / "ground.rad",
                ]
            },
            "windows": {
                "upper_glass": {
                    "file": objects_dir / "upper_glass.rad",
                    "matrix_name": "blinds30",
                },
                "lower_glass": {
                    "file": objects_dir / "lower_glass.rad",
                    "matrix_name": "blinds30",
                },
            },
            "materials": {
                "files": [objects_dir / "materials.mat"],
                "matrices": {
                    "blinds30": {"matrix_file": resources_dir / "blinds30.xml"}
                },
            },
            "sensors": {
                "wpi": {"file": resources_dir / "grid.txt"},
                "view1": {
                    "data": [[17, 5, 4, 1, 0, 0]],
                },
            },
            "views": {
                "view1": {
                    "file": resources_dir / "v1a.vf",
                    "xres": 16,
                    "yres": 16,
                }
            },
            "surfaces": {},
        },
    }


@pytest.fixture
def scene(objects_dir):
    return SceneConfig(
        files=[
            objects_dir / "walls.rad",
            objects_dir / "ceiling.rad",
            objects_dir / "floor.rad",
            objects_dir / "ground.rad",
        ]
    )


@pytest.fixture
def window_1(objects_dir):
    return WindowConfig(
        file=objects_dir / "upper_glass.rad",
        matrix_name="blinds30",
    )


@pytest.fixture
def window_2(objects_dir):
    return WindowConfig(
        file=objects_dir / "upper_glass.rad",
        # matrix_name="blinds30",
    )


@pytest.fixture
def materials(resources_dir, objects_dir):
    return MaterialConfig(
        files=[objects_dir / "materials.mat"],
        matrices={"blinds30": {"matrix_file": resources_dir / "blinds30.xml"}},
    )


@pytest.fixture
def wpi(resources_dir):
    return SensorConfig(
        file=resources_dir / "grid.txt",
    )


@pytest.fixture
def sensor_view_1():
    return SensorConfig(
        data=[[17, 5, 4, 1, 0, 0]],
    )


@pytest.fixture
def view_1(resources_dir):
    return ViewConfig(
        file=resources_dir / "v1a.vf",
        xres=16,
        yres=16,
    )


def test_model1(scene, window_1, materials, wpi, sensor_view_1, view_1):
    model = Model(
        scene=scene,
        windows={"window_1": window_1},
        materials=materials,
        sensors={"wpi": wpi, "view1": sensor_view_1},
        views={"view_1": view_1},
    )
    assert model.scene.files == scene.files
    assert model.windows["window_1"].file == window_1.file
    assert model.windows["window_1"].matrix_name == window_1.matrix_name
    assert model.materials.files == materials.files
    assert (
        model.materials.matrices["blinds30"].matrix_file
        == materials.matrices["blinds30"].matrix_file
    )
    assert model.windows["window_1"].matrix_name in model.materials.matrices
    assert model.sensors["wpi"].file == wpi.file
    assert model.sensors["view_1"].data == sensor_view_1.data
    assert model.views["view_1"].file == view_1.file
    assert model.views["view_1"].xres == view_1.xres
    assert model.views["view_1"].yres == view_1.yres


def test_model2(materials, wpi, view_1):
    # auto-generate view_1 in sensors from view_1 in views
    model = Model(
        materials=materials,
        sensors={"wpi": wpi},
        views={"view_1": view_1},
    )
    assert "view_1" in model.sensors
    assert model.sensors["view_1"].data == [
        model.views["view_1"].view.position + model.views["view_1"].view.direction
    ]
    assert isinstance(model.scene, SceneConfig)
    assert isinstance(model.windows, dict)
    assert model.scene.files == []
    assert model.scene.bytes == b""
    assert model.windows == {}


def test_model3(scene, window_1, materials, wpi, view_1):
    # same name view and sensor but different position and direction
    sensor_view_2 = SensorConfig(
        data=[[1, 5, 4, 1, 0, 0]],
    )

    with pytest.raises(ValueError):
        Model(
            scene=scene,
            windows={"window_1": window_1},
            materials=materials,
            sensors={"wpi": wpi, "view_1": sensor_view_2},
            views={"view_1": view_1},
        )


def test_model4(objects_dir, scene, window_1, materials, wpi, sensor_view_1, view_1):
    # window matrix name not in materials
    materials = MaterialConfig(files=[objects_dir / "materials.mat"])

    with pytest.raises(ValueError):
        Model(
            scene=scene,
            windows={"window_1": window_1},
            materials=materials,
            sensors={"wpi": wpi, "view_1": sensor_view_1},
            views={"view_1": view_1},
        )


def test_no_sensors_views_surfaces_specified(scene, window_1, materials):
    settings = Settings()
    model = Model(
        scene=scene,
        windows={"window_1": window_1},
        materials=materials,
    )
    with pytest.raises(ValueError):
        WorkflowConfig(settings, model)


def test_windows_not_specified_for_3phase_or_5phase_method(
    scene, materials, wpi, sensor_view_1, view_1
):
    settings = Settings()
    model = Model(
        scene=scene,
        # windows={"window_1": window_1},
        materials=materials,
        sensors={"wpi": wpi},
        views={"view_1": view_1},
    )
    with pytest.raises(ValueError):
        WorkflowConfig(settings, model)


def test_three_phase2(scene, window_2, materials, wpi, sensor_view_1, view_1):
    model = Model(
        scene=scene,
        windows={"window_1": window_2},  # window_2 has no matrix_name
        materials=materials,
        sensors={"wpi": wpi, "view_1": sensor_view_1},
        views={"view_1": view_1},
    )
    settings = Settings()

    cfg = WorkflowConfig(settings, model)
    workflow = ThreePhaseMethod(cfg)
    workflow.generate_matrices(view_matrices=False)
    a = workflow.calculate_sensor(
        "view_1",
        {"window_1": "blinds30"},  # blinds30 is the matrix_name
        datetime(2023, 1, 1, 12),
        800,
        100,
    )
    assert a.shape == (1, 1)


def test_two_phase(cfg):
    time = datetime(2023, 1, 1, 12)
    dni = 800
    dhi = 100
    config = WorkflowConfig.from_dict(cfg)
    with TwoPhaseMethod(config) as workflow:
        workflow.generate_matrices()
        res = workflow.calculate_sensor("wpi", time, dni, dhi)
    assert res.shape == (195, 1)


def test_three_phase(cfg, resources_dir):
    time = datetime(2023, 1, 1, 12)
    dni = 800
    dhi = 100
    config = WorkflowConfig.from_dict(cfg)
    blind_prim = pr.Primitive(
        "void",
        "aBSDF",
        "blinds30",
        [str(resources_dir / "blinds30.xml"), "0", "0", "1", "."],
        [],
    )
    config.model.materials.glazing_materials = {"blinds30": blind_prim}
    with ThreePhaseMethod(config) as workflow:
        workflow.generate_matrices(view_matrices=False)
        workflow.calculate_sensor(
            "wpi",
            {"upper_glass": "blinds30", "lower_glass": "blinds30"},
            time,
            dni,
            dhi,
        )
        res = workflow.calculate_edgps(
            "view1",
            {"upper_glass": "blinds30", "lower_glass": "blinds30"},
            time,
            dni,
            dhi,
        )
        res = workflow.calculate_sensor_from_wea("wpi")


def test_eprad_threephase(resources_dir):
    """
    Integration test for ThreePhaseMethod using EnergyPlusModel and GlazingSystem
    """
    view_path = resources_dir / "view1.vf"
    clear_glass_path = resources_dir / "CLEAR_3.DAT"
    product_7406_path = resources_dir / "igsdb_product_7406.json"
    shade_bsdf_path = resources_dir / "ec60.xml"

    epmodel = load_energyplus_model(ref_models["medium_office"])
    gs_ec60 = create_glazing_system(
        name="ec60",
        layers=[product_7406_path, clear_glass_path],
        gaps=[Gap([Gas("air", 0.1), Gas("argon", 0.9)], 0.0127)],
    )
    epmodel.add_glazing_system(gs_ec60)
    rad_models = epmodel_to_radmodel(
        epmodel, epw_file=weather_files["usa_ca_san_francisco"]
    )
    zone = "Perimeter_bot_ZN_1"
    zone_dict = rad_models[zone]
    zone_dict["model"]["views"]["view1"] = {
        "file": view_path,
        "xres": 16,
        "yres": 16,
    }
    zone_dict["model"]["sensors"]["view1"] = {
        "data": [[6.0, 7.0, 0.76, 0.0, -1.0, 0.0]]
    }
    zone_dict["model"]["materials"]["matrices"] = {
        "ec60": {"matrix_file": shade_bsdf_path}
    }
    zone_dict["model"]["surfaces"] = {}
    rad_cfg = WorkflowConfig.from_dict(zone_dict)
    rad_cfg.settings.sensor_window_matrix = ["-ab", "0"]
    rad_cfg.settings.view_window_matrix = ["-ab", "0"]
    rad_cfg.settings.daylight_matrix = ["-ab", "0"]
    with ThreePhaseMethod(rad_cfg) as rad_workflow:
        rad_workflow.generate_matrices(view_matrices=False)
        dni = 800
        dhi = 100
        dt = datetime(2023, 1, 1, 12)
        edgps, ev = rad_workflow.calculate_edgps(
            view="view1",
            bsdf={f"{zone}_Wall_South_Window": "ec60"},
            time=dt,
            dni=dni,
            dhi=dhi,
            ambient_bounce=1,
        )

    assert "view1" in rad_workflow.view_senders
    assert rad_workflow.view_senders["view1"].view.vtype == "a"
    assert rad_workflow.view_senders["view1"].view.position == [6.0, 7.0, 0.76]
    assert rad_workflow.view_senders["view1"].view.direction == [0.0, -1.0, 0.0]
    assert rad_workflow.view_senders["view1"].view.horiz == 180
    assert rad_workflow.view_senders["view1"].view.vert == 180
    assert rad_workflow.view_senders["view1"].xres == 16

    assert list(rad_workflow.daylight_matrices.values())[0].array.shape == (
        145,
        146,
        3,
    )
    assert (
        list(rad_workflow.sensor_window_matrices.values())[0].ncols == [145]
        and list(rad_workflow.sensor_window_matrices.values())[0].ncomp == 3
    )
    assert edgps >= 0 and edgps <= 1
